@@ -18,18 +18,42 @@
 #include <netlink/genl/ctrl.h>  
 #include <netlink/msg.h>
 #include <netlink/attr.h>
-#include <linux/nl80211.h>
 
+#include "nl80211.h"
 #include "iw.h"
 #include "version.h"
 
-int debug = 0;
+#ifndef CONFIG_LIBNL20
+/* libnl 2.0 compatibility code */
+
+static inline struct nl_handle *nl_socket_alloc(void)
+{
+	return nl_handle_alloc();
+}
+
+static inline void nl_socket_free(struct nl_handle *h)
+{
+	nl_handle_destroy(h);
+}
+
+static inline int __genl_ctrl_alloc_cache(struct nl_handle *h, struct nl_cache **cache)
+{
+	struct nl_cache *tmp = genl_ctrl_alloc_cache(h);
+	if (!tmp)
+		return -ENOMEM;
+	*cache = tmp;
+	return 0;
+}
+#define genl_ctrl_alloc_cache __genl_ctrl_alloc_cache
+#endif /* CONFIG_LIBNL20 */
+
+static int debug = 0;
 
 static int nl80211_init(struct nl80211_state *state)
 {
 	int err;
 
-	state->nl_handle = nl_handle_alloc();
+	state->nl_handle = nl_socket_alloc();
 	if (!state->nl_handle) {
 		fprintf(stderr, "Failed to allocate netlink handle.\n");
 		return -ENOMEM;
@@ -41,8 +65,7 @@ static int nl80211_init(struct nl80211_state *state)
 		goto out_handle_destroy;
 	}
 
-	state->nl_cache = genl_ctrl_alloc_cache(state->nl_handle);
-	if (!state->nl_cache) {
+	if (genl_ctrl_alloc_cache(state->nl_handle, &state->nl_cache)) {
 		fprintf(stderr, "Failed to allocate generic netlink cache.\n");
 		err = -ENOMEM;
 		goto out_handle_destroy;
@@ -60,7 +83,7 @@ static int nl80211_init(struct nl80211_state *state)
  out_cache_free:
 	nl_cache_free(state->nl_cache);
  out_handle_destroy:
-	nl_handle_destroy(state->nl_handle);
+	nl_socket_free(state->nl_handle);
 	return err;
 }
 
@@ -68,9 +91,14 @@ static void nl80211_cleanup(struct nl80211_state *state)
 {
 	genl_family_put(state->nl80211);
 	nl_cache_free(state->nl_cache);
-	nl_handle_destroy(state->nl_handle);
+	nl_socket_free(state->nl_handle);
 }
 
+__COMMAND(NULL, NULL, NULL, 0, 0, 0, CIB_NONE, NULL);
+__COMMAND(NULL, NULL, NULL, 1, 0, 0, CIB_NONE, NULL);
+
+static int cmd_size;
+
 static void usage(const char *argv0)
 {
 	struct cmd *cmd;
@@ -80,7 +108,12 @@ static void usage(const char *argv0)
 	fprintf(stderr, "\t--debug\t\tenable netlink debugging\n");
 	fprintf(stderr, "\t--version\tshow version\n");
 	fprintf(stderr, "Commands:\n");
-	for (cmd = &__start___cmd; cmd < &__stop___cmd; cmd++) {
+	fprintf(stderr, "\thelp\n");
+	fprintf(stderr, "\tevent\n");
+	for (cmd = &__start___cmd; cmd < &__stop___cmd;
+	     cmd = (struct cmd *)((char *)cmd + cmd_size)) {
+		if (!cmd->handler || cmd->hidden)
+			continue;
 		switch (cmd->idby) {
 		case CIB_NONE:
 			fprintf(stderr, "\t");
@@ -133,6 +166,8 @@ static int error_handler(struct sockaddr_nl *nla, struct nlmsgerr *err,
 
 static int finish_handler(struct nl_msg *msg, void *arg)
 {
+	int *ret = arg;
+	*ret = 0;
 	return NL_SKIP;
 }
 
@@ -176,7 +211,10 @@ static int handle_cmd(struct nl80211_state *state,
 	argc--;
 	argv++;
 
-	for (cmd = &__start___cmd; cmd < &__stop___cmd; cmd++) {
+	for (cmd = &__start___cmd; cmd < &__stop___cmd;
+	     cmd = (struct cmd *)((char *)cmd + cmd_size)) {
+		if (!cmd->handler)
+			continue;
 		if (cmd->idby != idby)
 			continue;
 		if (cmd->section) {
@@ -199,7 +237,7 @@ static int handle_cmd(struct nl80211_state *state,
 		break;
 	}
 
-	if (cmd == &__stop___cmd)
+	if (cmd >= &__stop___cmd)
 		return 1;
 
 	msg = nlmsg_alloc();
@@ -237,11 +275,14 @@ static int handle_cmd(struct nl80211_state *state,
 	if (err < 0)
 		goto out;
 
+	err = 1;
+
 	nl_cb_err(cb, NL_CB_CUSTOM, error_handler, &err);
-	nl_cb_set(cb, NL_CB_FINISH, NL_CB_CUSTOM, finish_handler, NULL);
+	nl_cb_set(cb, NL_CB_FINISH, NL_CB_CUSTOM, finish_handler, &err);
 	nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &err);
 
-	nl_recvmsgs(state->nl_handle, cb);
+	while (err > 0)
+		nl_recvmsgs(state->nl_handle, cb);
  out:
 	nl_cb_put(cb);
  out_free_msg:
@@ -252,16 +293,73 @@ static int handle_cmd(struct nl80211_state *state,
 	return 2;
 }
 
+static int no_seq_check(struct nl_msg *msg, void *arg)
+{
+	return NL_OK;
+}
+
+static int print_event(struct nl_msg *msg, void *arg)
+{
+	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
+	struct nlattr *tb[NL80211_ATTR_MAX + 1];
+
+	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+		  genlmsg_attrlen(gnlh, 0), NULL);
+                          
+	switch (gnlh->cmd) {
+	case NL80211_CMD_NEW_WIPHY:
+		printf("wiphy rename: phy #%d to %s\n",
+		       nla_get_u32(tb[NL80211_ATTR_WIPHY]),
+		       nla_get_string(tb[NL80211_ATTR_WIPHY_NAME]));
+		break;
+	default:
+		printf("unknown event: %d\n", gnlh->cmd);
+		break;
+	}
+
+	return NL_SKIP;
+}
+
+static int listen_events(struct nl80211_state *state,
+			 int argc, char **argv)
+{
+	int mcid, ret;
+	struct nl_cb *cb = nl_cb_alloc(debug ? NL_CB_DEBUG : NL_CB_DEFAULT);
+
+	if (!cb) {
+		fprintf(stderr, "failed to allocate netlink callbacks\n");
+		return -ENOMEM;
+	}
+
+	mcid = nl_get_multicast_id(state->nl_handle, "nl80211", "config");
+	if (mcid < 0)
+		return mcid;
+
+	ret = nl_socket_add_membership(state->nl_handle, mcid);
+	if (ret)
+		return ret;
+	
+	/* no sequence checking for multicast messages */
+	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, no_seq_check, NULL);
+	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, print_event, NULL);
+
+	while (1)
+		nl_recvmsgs(state->nl_handle, cb);
+
+	nl_cb_put(cb);
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	struct nl80211_state nlstate;
 	int err;
 	const char *argv0;
 
-	err = nl80211_init(&nlstate);
-	if (err)
-		return 1;
-
+	/* calculate command size including padding */
+	cmd_size = abs((long)&__cmd_NULL_1_CIB_NONE_0
+	             - (long)&__cmd_NULL_0_CIB_NONE_0);
 	/* strip off self */
 	argc--;
 	argv0 = *argv++;
@@ -279,10 +377,16 @@ int main(int argc, char **argv)
 
 	if (argc == 0 || strcmp(*argv, "help") == 0) {
 		usage(argv0);
-		goto out;
+		return 0;
 	}
 
-	if (strcmp(*argv, "dev") == 0) {
+	err = nl80211_init(&nlstate);
+	if (err)
+		return 1;
+
+	if (strcmp(*argv, "event") == 0) {
+		err = listen_events(&nlstate, argc, argv);
+	} else if (strcmp(*argv, "dev") == 0) {
 		argc--;
 		argv++;
 		err = handle_cmd(&nlstate, CIB_NETDEV, argc, argv);
@@ -298,7 +402,6 @@ int main(int argc, char **argv)
 	if (err < 0)
 		fprintf(stderr, "command failed: %s (%d)\n", strerror(-err), err);
 
- out:
 	nl80211_cleanup(&nlstate);
 
 	return err;
